(* $Id: godi_script.mli 1094 2012-08-13 00:13:54Z gerd $ *)

(** This module contains library functions for GODI configuration scripts
 * written in OCaml. This is mainly intended for the conf-foo packages.
 *
 * In order to activate such a script, define the following variables in
 * the driver Makefile of conf-foo:
 *
 * - [CONF_SCRIPT]: The name of the script, usually [configure.gcs]
 *   (gcs = GODI configuration script). The suffix "gcs" enables a special
 *   calling convention that automatically finds the O'Caml interpreter
 *   and that loads the libraries [str] and [unix], and, of course,
 *   [godi_script].
 * - [CONF_SCRIPT_ARGS]: Arguments to be passed to the script.
 *
 * A number of environment variables are automatically passed to this
 * script:
 *
 * - [LOCALBASE]: The GODI prefix of this installation
 * - [CC]: The name of the C compiler
 * - [CFLAGS]: General flags for C compiling
 * - [LDFLAGS]: General flags for linking
 * - [CREATE_SHLIBS]: Whether C libraries must be shared ("yes") or not
 * - [SHLIB_TYPE]: The type of shared libraries used by the OS
 * - [ELF_RPATH]: Whether to use RPATH at all ("yes") or not
 * - [ELF_RPATH_FLAG]: The name of the RPATH flag, if any.
 * - [SEARCH_LIBS]: The list of standard library locations
 * - [MAKE]: The name of the [make] program to use for upstream Makefiles
 * - Furthermore, a number of programs like [RM], [CP], ...
 *
*)

(** {1 Primitives} *)

val runcmd : string -> int
  (** Starts the command, waits for its completion, and returns the exit
   * code. Similar to [Sys.command], but preferable. 
   *)

val runcmdf : ('a, unit, string, int) format4 -> 'a
  (** Same as [runcmd], but accepts a [printf]-like format string. *)

val findcmd : ?path:string list -> string -> string
  (** Returns the absolute location of a command, or raises [Not_found].
   * The command is searched in the locations enumerated by the environment
   * variable [PATH], by default, or in the locations passed by the
   * [path] argument.
   *)

val system_path : unit -> string list
  (** Returns the locations enumerated by the environment
   * variable [PATH]
   *)

val search_libs_path : unit -> string list
  (** Returns the (command) locations enumerated by the environment
   * variable [SEARCH_LIBS]
   *)

val read_file : string -> string
  (** Reads the file into memory, and returns it as string *)

val write_file : string -> string -> unit
  (** Writes to the file whose name is passed as first argument the
   * 0contents passed as second argument.
   *)

val read_from_cmd : string -> (int * string)
  (** Starts the command, reads its output (stdout), waits for its
   * completion, and returns both the exit code and the output.
   *)

val read_from_cmdf : ('a, unit, string, int * string) format4 -> 'a
  (** Same as [read_from_cmd], but accepts a [printf]-like format string. *)

val env : string -> string
  (** Returns the named environment variable, or [""] if the variable does
   * not exist.
   *)

val env_req : string -> string
  (** Returns the named environment variable, or fails if the variable
   * does not exist {b or is empty}.
   *)

val env_opt : string -> string option
  (** Returns the named environment variable, or [None] if the variable
   * is empty or does not exist.
   *)


(** {1 Commandlets} *)

(** A "commandlet" is a command function that takes a state variable as
 * input and that returns the possibly modified state variable as output.
 * The state variable contains the exit code of the last executed shell
 * command, and optionally a further result value.
 *
 * Remember that the exit code 0 means that the command was successful.
 *)

type 'a cmdstate
  (** The state with further values of type ['a] *)

val code : 'a cmdstate -> int
  (** The current exit code *)

val result : 'a cmdstate -> 'a option
  (** The current result value *)

val return_code : int -> 'a cmdstate -> 'a cmdstate
  (** Returns the modified state where the code is set to the int *)

val return_result : 'a option -> 'a cmdstate -> 'a cmdstate
  (** Returns the modified state where the result is set *)

val initial_state : unit -> 'a cmdstate
  (** The initial state has code 0, and no result value *)


type 'a cmdlet = 'a cmdstate -> 'a cmdstate
  (** A commandlet is a function taking state and returning state *)

val cmd : string -> 'a cmdlet
  (** The commandlet that executes the passed command (as string). The
   * exit code is set in the resulting state.
   *)

val cmdf : ('a, unit, string, 'b cmdlet) format4 -> 'a
  (** Same as [cmd] but accepts a [printf]-like format string *)

val cmd_output : string -> string cmdlet
  (** The commandlet that executes the passed command (as string), and
   * reads the output of the command (stdout). The exit code is set in the
   * resulting state, and the output is returned as result value.
   *)

val cmdf_output : ('a, unit, string, string cmdlet) format4 -> 'a
  (** Same as [cmd_output] but accepts a [printf]-like format string *)

val set_code : int -> 'a cmdlet
  (** The commandlet that sets the exit code to the passed integer *)

val set_code_from : ('a cmdstate -> int) -> 'a cmdlet
  (** The commandlet that calls the function, and sets the exit code
   * from the integer result of the function.
   *)

val set_bool_code : bool -> 'a cmdlet
  (** The commandlet that sets the exit code to 0 if the boolean is
   * [true], and to 1 if the boolean is [false].
   *)

val set_bool_code_from : ('a cmdstate -> bool) -> 'a cmdlet
  (** The commandlet that calls the function, and sets the exit code
   * from the boolean result of the function (as [set_code_bool]).
   *
   * This function is useful to call non-commandlets from a sequence
   * of commandlets. For instace, to just print messages:
   *
   * {[ 
   * cmd1 &- 
   * set_bool_code_from (fun _ -> printf "Here I am\n"; true) &-
   * cmd2 
   * ]}
   *)

val ( &- ) : 'a cmdlet -> 'a cmdlet -> 'a cmdlet
  (** Conditional sequencing: If the first commandlet returns the exit
   * code 0, the second commandlet is also executed (like the "&&" shell
   * operator).
   *)

val ( |- ) : 'a cmdlet -> 'a cmdlet -> 'a cmdlet
  (** Conditional sequencing: If the first commandlet returns a non-zero exit
   * code, the second commandlet is also executed (like the "||" shell
   * operator).
   *)

val ignore_code : 'a cmdlet -> 'a cmdlet
  (** Executes the commandlet, and ignores the exit code, i.e. sets it
   * always to 0.
   *)

val eval : 'a cmdlet -> 'a option
  (** Evaluates a commandlet, and returns the result value. Note that
   * [None] is always returned if the exit code is non-zero, regardless of
   * whether there is a result value or not.
   *)

val eval_test : 'a cmdlet -> bool
  (** Evaluates a commandlet, and returns [true] if the exit code is 0,
   * and [false] if the exit code is non-zero.
   *)


(** Example: This piece of code returns the file listing of "/a" if
 * readable, or otherwise the file listing of "/b":
 *
 * {[ 
 * eval (cmd_output "ls /a" |- cmd_output "ls /b")
 * ]}
 *
 * Loops should be programmed with recursion, or by folding:
 *
 * {[
 * eval
 *   (List.fold_left
 *     (fun acc x ->
 *        acc |- cmdf_output "ls %s" (Filename.quote x))
 *     (set_bool_code true)
 *     [ "/a"; "/b"; "/c" ]
 *   )
 * ]}
 *)


(** {1 Logging} *)

(** Logging outputs to "script.log" in the cwd. *)

val log : string -> unit
  (** Appends the string to the log file. *)

val logf : ('b, unit, string, unit) format4 -> 'b
  (** Same as [log], but accepts a [printf]-like format string *)


(** {1 The conf-*.mk files} *)

(** In general, the conf-*.mk files have [make] syntax. However, these
 * files should only contain [name=value] settings, and not more.
 * The values should not contain any references to other variables
 * ("$\{name\}"), and the dollar character must be written [$$].
 *)

val print_conf_file : string -> (string * string) list -> unit
  (** Writes to [file] the list of variable settings *)

val parse_conf_file : string -> (string * string) list
  (** Parses [file] *)

(** {1 Finding libraries} *)

type lang_env =
    [ `C of c_env ]
  (** Environment to compile and link programs for a certain language. *)

and c_env = 
  { c_env_file : (string * string) option; 
      (** The file this record was parsed from. The first string is the
        * absolute file name, and the second string is the name prefix used
        * in this file
        *)
    c_incdirs : string list;  (** List of directories for "#include" *)
    c_libdirs : string list;  (** List of library directories *)
    c_libs : string list;     (** Names of libraries (w/o "lib" and suffix) *)
    c_cflags : string list;   (** Further flags for C compiling *)
    c_ldflags : string list;  (** Further flags for linking *)
    c_elf_rpath : bool;       (** Whether [c_libdirs] must be added to the
			       * runtime lookup path
			       *)
    c_config_script : (string * string * string) option; (** The name of a
                                  foreign config script, plus the args for
                                  getting cflags, plus the args for
                                  getting ldflags *)
    c_godi_deps : string list;  (** List of additional GODI dependencies to
				 * output
				 *)
    c_requirements : c_env list; (** Requirements to link test programs *)
  }
  (** The environment to compile and link the mentioned libraries.
   * The [c_incdirs] can be translated to further C compiler flags;
   * the [c_libdirs] and [c_libs] can be translated to further linker
   * flags. The [c_cflags] and [c_ldflags] are additional flags.
   * It is also allowed to only fill [c_cflags] and [c_ldflags] and let
   * the other components empty.
   *
   * The [c_config_script] is the name of another script that can be called
   * to get the C compiler and linker flags (e.g. pkg-config scripts).
   * The script is specified as triple [(script,cflags_args,ldflags_args)]
   * where [script] is the path to the script (e.g. /usr/bin/pkg-config),
   * and [cflags_args] are the arguments to pass for getting the CFLAGS,
   * e.g. "--cflags foo", and [ldflags_args] are the arguments for getting
   * LDFLAGS, e.g. "--libs foo".
   *
   * The [c_godi_deps] are dependencies to GODI packages that are required
   * at run-time to make the C library available. (These are usually
   * base-foo packages. Configuration packages must not be put into this
   * list!)
   *
   * The [c_requirements] lists environments that are necessary to compile
   * and link programs, but that are not merged into the regular output. 
   * For instance,
   * many graphics libraries depend on the X11 library.
   *
   * A note about shell quoting: The above lists of flags should not
   * contain any shell metacharacters such that quoting is not necessary
   * at all. As the flags are processed by further scripts it is expected
   * that quoting will be taken wrong anyway, so it is the best strategy
   * to avoid quoting.
   *)

val empty_c_env : c_env
  (** A [c_env] where the lists are empty, the options are [None], and the
   * boolean fields are [false].
   *)

val c_env_from_config_script :
      ?arg_cflags:string ->
      ?arg_ldflags:string ->
      ?args:string list ->
      script_name:string ->
      unit ->
	c_env
  (** Calls the script [script_name] (which must be an absolute file name)
   * and extracts the [c_env] from it. The C compiler flags are obtained
   * by calling the script with argument [arg_cflags], and the linker
   * flags are obtained by calling the script with argument [arg_ldflags].
   *
   * The default for [arg_cflags] is ["--cflags"], and the default
   * for [arg_ldflags] is ["--libs"]. (These work in most cases.)
   *
   * In [args] one can pass further arguments.
   *
   * The function fails if the configure script cannot be called, or
   * if it does not exit with code 0.
   *)


val create_test_whether_c_function_exists : string -> string
  (** Input is the name of a C function. The returned string is a 
   * C program that can only be compiled and linked if the C function exists.
   * The program returns the exit code 0.
   *)

val c_cflags : ?with_requirements:bool -> c_env -> string list
  (** Returns the collected flags to C-compile.
   * If [with_requirements=true], the flags mentioned in [c_requirements]
   * are also returned (default: [false]).
   *)

val c_ldflags : ?with_requirements:bool -> c_env -> string list
  (** Returns the collected flags to link C programs.
   * If [with_requirements=true], the flags mentioned in [c_requirements]
   * are also returned (default: [false]).
   *)

val c_compiler_name : unit -> string
  (** Returns the command name of the C compiler *)

val c_compile_test : c_env -> string -> string cmdlet
  (** The string must contain the C program to be used as test. The
   * program is compiled and linked using the C environment. The result
   * stored into the commandlet state is the name of the executable.
   *)

val run_test : string cmdlet
  (** Runs the executable whose name is stored in the state. The exit
   * code is set, and the new result is the output of the program.
   *)

val find_c_library :
      ?godi_deps:string list ->
      ?pref_incdir:string ->
      ?pref_libdir:string ->
      ?cflags:string list ->
      ?ldflags:string list ->
      ?inc_name:string ->
      ?requirements:c_env list ->
      ?locations:string list ->
      ?frameworks:string list ->
      libs:string list ->
      test:(c_env -> c_env option) ->
      unit ->
	c_env cmdlet
  (** Searches a C library. First, the preferred locations [pref_incdir]
   * and [pref_libdir] are tried. The second choice are the frameworks
   * (Mac OS X only). If this is not successful, too, searching continues
   * with the locations found in the environment variable [SEARCH_LIBS]
   * (by default, unless overriden by [locations]).
   * 
   * A C library is found if:
   * - The include file [inc_name] is found (if passed), and
   * - the [test] returns a non-[None] value
   *
   * If found, the return value of [test] is the result of the
   * commandlet.
   *
   * The C environment passed to the [test] function is properly set
   * up:
   * - [c_incdirs] contains the tested include location
   * - [c_libdirs] contains the tested library location
   * - [c_libs] is [libs]
   * - [c_cflags] is [cflags]
   * - [c_ldflags] is [ldflags], optionally enriched by the necessary
   *   framework flags
   * - [c_elf_rpath] is tried 
   * - [c_config_script] is always [None]
   * - [c_godi_deps] is set to [godi_deps], but only when the preferred
   *   location is tried
   *
   * @param frameworks If non-empty, and if the system is Mac OS X, this
   * is a list of frameworks to try.
   *)

val parse_c_result : file:string -> prefix:string -> unit -> c_env
  (** Parses the C environment found in [file] where the variables
   * use [prefix].
   *
   * The syntax written by [print_c_result] is recognised plus a number
   * of legacy notations.
   *)

val print_c_result : ?vars:(string*string)list -> 
                     file:string -> prefix:string -> c_env -> unit
  (** Prints the C environment to the [file] and to the screen.
   * The [prefix] is prepended to the variables in [file].
   * Example: [print_c_result ~file:"conf-foo.mk" ~prefix:"CONF_FOO" ce]
   *
   * The component [c_env_file] is ignored (except in [c_requirements]
   * where it is output to the [<prefix>_REQS] variable).
   *
   * The following standard variable names are used in the printed [file]:
   * - [<prefix>_INCDIRS]: The include directories. /usr/include is omitted
   *   from this list.
   * - [<prefix>_LIBDIRS]: The library directories. /usr/lib is omitted
   *   from this list.
   * - [<prefix>_LIBNAMES]: The library names.
   * - [<prefix>_CFLAGS]: The additional flags for C compiling
   * - [<prefix>_LDFLAGS]: The additional flags for linking
   * - [<prefix>_NEED_RPATH]: Whether the RPATH flag is needed ("yes") or not
   *   ("no")
   * - [<prefix>_TOTAL_CFLAGS]: The combined flags for C compiling from
   *   both [INCDIRS] and [CFLAGS]. Flags from required libraries are omitted.
   * - [<prefix>_TOTAL_LDFLAGS]: The combined flags for linking from
   *   all of [LIBDIRS], [LIBS], [LDFLAGS], and [NEED_RPATH]. Flags from
   *   required libraries are omitted.
   * - [<prefix>_SCRIPT]: The name of the external config script
   * - [<prefix>_DEPENDS]: The additional GODI dependencies
   * - [<prefix>_REQS]: The requirements, if bound to files (to be used
   *   by nested configuration packages)
   *
   * The [vars] are appended to the file, if passed. Useful to output
   * further non-standard variables.
   *)


val main_c_finder :
      ?godi_deps:(string list ref) ->
      ?pref_incdir:(string option ref) ->
      ?pref_libdir:(string option ref) ->
      ?cflags:(string list ref) ->
      ?ldflags:(string list ref) ->
      ?parse_args:((Arg.key * Arg.spec * Arg.doc) list) ->
      unit -> unit
  (** A configurable main program for the case that the library
   * is searched by [find_c_library].
   *
   * The arguments [godi_deps],
   * [pref_incdir], [pref_libdir], [c_flags], and [ld_flags] can
   * be set by command-line if a string reference is passed to the
   * function.
   *
   * Example:
   * {[
   * let pref_incdir = ref None in
   * let pref_libdir = ref None in
   * main_c_finder ~pref_incdir ~pref_libdir ();
   * let c_opt = eval
   *               (find_c_library 
   *                  ?pref_incdir:!pref_incdir 
   *                  ?pref_libdir:!pref_libdir 
   *                  ~libs:... ~test:... ()) in
   * match c_opt with
   *   None -> failwith "Cannot find library"
   * | Some c -> print_c_result ~file:"conf-foo.mk" ~prefix:"CONF_FOO" c
   * ]}
   *)

val main_config_finder :
      script_basename:string ->
      unit -> string
  (** A configurable main program for the case that a configure script
   * is used ([c_env_from_config_script]).
   *
   * In [script_basename] one must pass the name of the script to call,
   * without directory, e.g. ["foo-config"]. The script is looked up
   * in the [SEARCH_LIBS] and [PATH] locations. A command-line parameter
   * may be used to override the location.
   *
   * The absolute path of the script is returned as string.
   *
   * Whole example:
   * {[
   * let test ce = ... ;;
   * let cmd = main_config_finder ~script_basename:"foo-config" () ;;
   * let ce = c_env_from_config_script ~script_name:cmd () ;;
   * match test ce with
   *   Some ce' ->
   *     print_c_result ~file:"conf-foo.mk" ~prefix:"CONF_FOO" ce'
   * | None ->
   *     failwith "Cannot find library"
   * ]}
   *)
